from apgl.predictors.AbstractSVMPredictor import AbstractSVMPredictor
from apgl.util.Parameter import Parameter
from apgl.util.Sampling import Sampling
from apgl.util.Evaluator import Evaluator
from apgl.util.Util import Util

import sys 
import numpy
import scipy
import scipy.sparse
import multiprocessing

def computeTestError(args):
    """
    Used in conjunction with the parallel model selection.
    """
    (trainX, trainY, testX, testY, C, gamma) = args

    svm = LibSVM('gaussian', gamma, C)
    svm.learnModel(trainX, trainY)
    predY = svm.predict(testX)

    return Evaluator.binaryError(predY, testY)

def computePenalisedError(args):
    """
    Used in conjunction with the parallel model selection.
    """
    (X, y, idx, C, gamma, Cv) = args
    penalty = 0
    folds = len(idx)
    svm = LibSVM('gaussian', gamma, C)

    for idxtr, idxts in idx:
        trainX, trainY = X[idxtr, :], y[idxtr]
        svm.learnModel(trainX, trainY)
        predY = svm.predict(X)
        predTrainY = svm.predict(trainX)
        penalty += Evaluator.binaryError(predY, y) - Evaluator.binaryError(predTrainY, trainY)

    penalty *= Cv/folds

    svm.learnModel(X, y)
    predY = svm.predict(X)
    return Evaluator.binaryError(predY, y) + penalty


class LibSVM(AbstractSVMPredictor):
    def __init__(self, kernel='linear', kp=0.1, C=1.0, cost=1.0):
        try:
            import svm
        except ImportError:
            raise 

        self.setKernel(kernel, kp)
        self.setC(C)
        self.setErrorCost(cost)
        self.setSvmType("C_SVC")
        self.setTermination(0.001)

        #Parameters used for model selection
        self.Cs = 2.0**numpy.arange(-10, 20, dtype=numpy.float)
        self.gammas = 2.0**numpy.arange(-10, 4, dtype=numpy.float)

    def setC(self, C):
        Parameter.checkFloat(C, 0.0, float('inf'))
        
        self.C = C 

    def setKernel(self,  kernel='linear', kernelParam=0.1):
        try:
            import svm
        except:
            raise 
        Parameter.checkString(kernel, ["linear", "gaussian", "polynomial"])
        kernelStr = ""

        if kernel == "linear":
            kernelStr = "-t 0"
        elif kernel == "polynomial":
            kernelStr = "-t 1 -d " + str(int(kernelParam))
        elif kernel == "gaussian":
            kernelStr = "-t 2 -g " + str(kernelParam)
        else:
            raise ValueError("Invalid kernel :" + str(kernel))

        self.kernelStr = kernelStr
        self.kernelParam = kernelParam

    def setSvmType(self, type):
        try:
            import svm
        except:
            raise 
        
        self.type = type

        if type == "C_SVC":
            typeStr = "-s 0"
        elif type == "Epsilon_SVR":
            typeStr = "-s 3"
        else:
            raise ValueError("Invalid SVM type :" + str(type))

        self.typeStr = typeStr

    def setErrorCost(self, errorCost):
        """
        The penalty on errors on positive labels. The penalty for negative labels
        is 1.
        """
        Parameter.checkFloat(errorCost, 0.0, float('inf'))
        self.errorCost = errorCost

    def setEpsilon(self, epsilon):
        """
        This is for backward compatibility only
        """
        self.setTermination(epsilon)

    def setTermination(self, epsilon):
        Parameter.checkFloat(epsilon, 0.0, 1.0)
        self.epsilon = epsilon

    def learnModel(self, X, y):
        try:
            import svmutil
        except:
            raise 

        if self.typeStr == "-s 0" and (not ((y == 1) + (y == -1)).all()):
            raise ValueError("Labels must be chosen from {-1, +1} for SVC")

        self.numFeatures = X.shape[1]

        cStr = "-c " + str(self.C)
        errorCostStr = "-w1 " + str(self.errorCost) + " -w1 1"
        termStr = "-p " + str(self.epsilon)
            
        paramStr = "-q " + self.typeStr + " " + termStr + " " + self.kernelStr + " " + " " + cStr + " " + errorCostStr
        self.model = svmutil.svm_train(y.tolist(), X.tolist(), paramStr)

    def classify(self, X):
        try:
            import svmutil
        except:
            raise
        
        if self.typeStr != "-s 0":
            raise ValueError("Cannot classify without correct svm_type")

        return self.predict(X, False)

    def predict(self, X, decisionValues=False): 
        """
        Classify a set of examples and return an array of predicted labels
        """
        try:
            import svmutil
        except:
            raise

        y = numpy.zeros(X.shape[0])

        from StringIO import StringIO
        backupStdout=sys.stdout
        sys.stdout=StringIO()
        predY, predAcc, decisionVals = svmutil.svm_predict(y.tolist(), X.tolist(), self.model)
        sys.stdout=backupStdout

        if not decisionValues:
            return numpy.array(predY)
        else:
            labels = self.model.get_labels()
            if labels[0] == -1:
                return numpy.array(predY), -numpy.array(decisionVals).ravel()
            else:
                return numpy.array(predY), numpy.array(decisionVals).ravel()
        
    def getWeights(self):
        """
        A method to get the weight vector and bias b. The decision function is made
        using the sign of numpy.dot(X, weights) + b.
        """
        try:
            import svm
        except:
            raise

        b = -self.model.rho.contents.value

        svCoef = self.model.get_sv_coef()
        sv = self.model.get_SV()

        #Convert support vectors to sparse matrix
        numSVs = len(sv)

        SVMatrix = scipy.sparse.csr_matrix((numSVs, self.numFeatures))
        i = 0

        for row in sv:
            for key, value in row.items():
                if key != -1:
                    SVMatrix[i, key] = value
            i += 1

        svCoefVector = scipy.sparse.csr_matrix((numSVs, numSVs))
        i = 0 
        for val in svCoef:
            svCoefVector[i, i] = val[0]
            i = i+1

        w = numpy.array((svCoefVector * SVMatrix).sum(0)).flatten()
        labels = self.model.get_labels()

        if labels[0] == -1:
            w = -w
            b = -b

        return w, b

    def loadParams(self, fileName):
        """
        Load a set of SVM parameters from a matlab file. 
        """
        try:
            matDict = scipy.io.loadmat(fileName)
        except IOError:
            raise

        if "C" in list(matDict.keys()):
            self.setC(float(matDict["C"][0]))

        if "kernel" in list(matDict.keys()):
            self.setKernel(str(matDict["kernel"][0]), float(matDict["kernelParamVal"][0])) 

        if "errorCost" in list(matDict.keys()):
            self.setErrorCost(float(matDict["errorCost"][0]))

        if "epsilon" in list(matDict.keys()):
            self.setTermination(float(matDict["epsilon"][0]))

        if "svmType" in list(matDict.keys()):
            self.setSvmType(str(matDict["svmType"][0]))

    def saveParams(self, fileName):
        raise NotImplementedError("This is not implemented")

    def getC(self):
        return self.C

    def getErrorCost(self):
        return self.errorCost 

    def getKernel(self):
        return self.kernelStr

    def getTermination(self):
        return self.epsilon

    def getSvmType(self):
        return self.svmTypeStr

    def getKernelParams(self):
        return self.kernelParam

    def __str__(self):
        cStr = "-c " + str(self.C)
        errorCostStr = "-w1 " + str(self.errorCost) + " -w1 1"
        termStr = "-p " + str(self.epsilon)
        paramStr = self.typeStr + " " + termStr + " " + self.kernelStr + " " + " " + cStr + " " + errorCostStr
        return "LibSVM: " + paramStr

    def getModel(self):
        return self.model

    def vfcvRbf(self, X, y, folds):
        """
        Basically perform cross validation model selection using the RBF kernel
        and then pick the best one. Using the best set of parameters train using
        the whole dataset. 
        """
        Parameter.checkInt(folds, 1, float('inf'))
        Parameter.checkClass(X, numpy.ndarray)
        Parameter.checkClass(y, numpy.ndarray)

        meanErrors = numpy.zeros((self.Cs.shape[0], self.gammas.shape[0]))
        stdErrors = numpy.zeros((self.Cs.shape[0], self.gammas.shape[0]))

        for i in range(self.Cs.shape[0]):
            for j in range(self.gammas.shape[0]):
                svm = LibSVM('gaussian', self.gammas[j], self.Cs[i])
                idx = Sampling.crossValidation(folds, y.shape[0])
                errors = svm.evaluateLearn(X, y, idx, svm.learnModel, svm.predict, Evaluator.binaryError, progress=True)
                meanErrors[i, j] = numpy.mean(errors)
                stdErrors[i, j] = numpy.std(errors)

        indC, indGamma = numpy.unravel_index(numpy.argmin(meanErrors), meanErrors.shape)
        bestC, bestGamma = self.Cs[indC], self.gammas[indGamma]
        svm = LibSVM('gaussian', bestGamma, bestC)
        svm.learnModel(X, y)

        return svm 

    def parallelVfcvRbf(self, X, y, folds):
        """
        Perform parallel cross validation model selection using the RBF kernel
        and then pick the best one. Using the best set of parameters train using
        the whole dataset.
        """
        Parameter.checkClass(X, numpy.ndarray)
        Parameter.checkClass(y, numpy.ndarray)
        Parameter.checkInt(folds, 1, float('inf'))

        meanErrors = numpy.zeros((self.Cs.shape[0], self.gammas.shape[0]))
        idx = Sampling.crossValidation(folds, y.shape[0])
        m = 0

        for trainInds, testInds in idx:
            Util.printConciseIteration(m, 1, folds, "CV Inner fold: ")
            trainX, trainY = X[trainInds, :], y[trainInds]
            testX, testY = X[testInds, :], y[testInds]

            paramList = []
            for i in range(self.Cs.shape[0]):
                for j in range(self.gammas.shape[0]):
                    paramList.append((trainX, trainY, testX, testY, self.Cs[i], self.gammas[j]))

            pool = multiprocessing.Pool()
            resultsIterator = pool.imap(computeTestError, paramList)
            pool.close()
            pool.join()

            for i in range(self.Cs.shape[0]):
                for j in range(self.gammas.shape[0]):
                    meanErrors[i, j] += resultsIterator.next()/float(folds)

            m += 1 

        indC, indGamma = numpy.unravel_index(numpy.argmin(meanErrors), meanErrors.shape)
        bestC, bestGamma = self.Cs[indC], self.gammas[indGamma]
        svm = LibSVM('gaussian', bestGamma, bestC)
        svm.learnModel(X, y)

        return svm 

    def vfPenRbf(self, X, y, folds, Cv):
        """
        Perform v fold penalisation model selection using the RBF kernel
        and then pick the best one. Using the best set of parameters train using
        the whole dataset. Cv is the control on the amount of penalisation.
        """
        Parameter.checkClass(X, numpy.ndarray)
        Parameter.checkClass(y, numpy.ndarray)
        Parameter.checkInt(folds, 1, float('inf'))
        Parameter.checkFloat(Cv, 0.0, float('inf'))

        meanErrors = numpy.zeros((self.Cs.shape[0], self.gammas.shape[0]))

        for i in range(self.Cs.shape[0]):
            for j in range(self.gammas.shape[0]):
                idx = Sampling.crossValidation(folds, y.shape[0])

                #Compute penalty here
                penalty = 0

                for idxtr, idxts in idx:
                    trainX, testX = X[idxtr, :], X[idxts, :]
                    trainY, testY = y[idxtr], y[idxts]
                    svm = LibSVM('gaussian', self.gammas[j], self.Cs[i])
                    svm.learnModel(trainX, trainY)
                    predY = svm.predict(X)
                    predTrainY = svm.predict(trainX)
                    penalty += Evaluator.binaryError(predY, y) - Evaluator.binaryError(predTrainY, trainY)

                penalty *= Cv/folds
                
                svm.learnModel(X, y)
                predY = svm.predict(X)
                meanErrors[i, j] = Evaluator.binaryError(predY, y) +  penalty

        indC, indGamma = numpy.unravel_index(numpy.argmin(meanErrors), meanErrors.shape)
        bestC, bestGamma = self.Cs[indC], self.gammas[indGamma]
        svm = LibSVM('gaussian', bestGamma, bestC)
        svm.learnModel(X, y)

        return svm 

    def parallelVfPenRbf(self, X, y, folds, Cv):
        """
        Perform v fold penalisation model selection using the RBF kernel
        and then pick the best one. Using the best set of parameters train using
        the whole dataset. Cv is the control on the amount of penalisation.
        """
        Parameter.checkClass(X, numpy.ndarray)
        Parameter.checkClass(y, numpy.ndarray)
        Parameter.checkInt(folds, 1, float('inf'))
        Parameter.checkFloat(Cv, 0.0, float('inf'))

        meanErrors = numpy.zeros((self.Cs.shape[0], self.gammas.shape[0]))
        idx = Sampling.crossValidation(folds, y.shape[0])
        paramList = []

        for i in range(self.Cs.shape[0]):
            for j in range(self.gammas.shape[0]):
                paramList.append((X, y, idx, self.Cs[i], self.gammas[j], Cv))

        pool = multiprocessing.Pool()
        resultsIterator = pool.imap(computePenalisedError, paramList)
        pool.close()
        pool.join()

        for i in range(self.Cs.shape[0]):
            for j in range(self.gammas.shape[0]):
                meanErrors[i, j] += resultsIterator.next()
    
        indC, indGamma = numpy.unravel_index(numpy.argmin(meanErrors), meanErrors.shape)
        bestC, bestGamma = self.Cs[indC], self.gammas[indGamma]
        svm = LibSVM('gaussian', bestGamma, bestC)
        svm.learnModel(X, y)

        return svm

